import shutil
import tempfile
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.management import call_command
from django.forms.models import modelform_factory
from django.test import TestCase
from django.test.utils import override_settings
import base64 as b64
import os
from webcam.utils import InvalidImageFormat, Base64Image
from webcam.tests.models import DBDemoModel, FSDemoModel


APPS = ['webcam.tests'] + list(settings.INSTALLED_APPS)


def override_app():
    return override_settings(INSTALLED_APPS=APPS, MEDIA_ROOT=tempfile.gettempdir())

HERE = os.path.dirname(__file__)
PICTURE_NAME = 'colosseo.jpg'
PICTURE_PATH = os.path.join(HERE, PICTURE_NAME)
PICTURE = file(PICTURE_PATH).read()


class CameraFieldTest(TestCase):
    def _pre_setup(self):
        with self.settings(INSTALLED_APPS=APPS):
            from django.core.management import call_command
            from django.db.models import loading

            loading.cache.loaded = False
            call_command('syncdb', verbosity=0)
        super(CameraFieldTest, self)._pre_setup()


class TestDBCameraField(CameraFieldTest):
    @override_app()
    def test_database(self):
        base64_picture = b64.encodestring(PICTURE)
        obj = DBDemoModel.objects.create(photo=Base64Image(base64_picture))
        obj.save()
        obj = DBDemoModel.objects.get(pk=obj.pk)
        self.assertEqual(obj.photo.size, 53472)
        self.assertEqual(len(obj.photo), 72235)
        self.assertEqual(obj.photo.width, 320)
        self.assertEqual(obj.photo.height, 200)
        self.assertTrue(obj.photo.image)

    @override_app()
    def test_empty_field(self):
        obj = DBDemoModel()
        obj.save()
        self.assertFalse(obj.photo)
        self.assertRaises(InvalidImageFormat, lambda: obj.photo.size)
        self.assertRaises(InvalidImageFormat, lambda: obj.photo.width)
        self.assertRaises(InvalidImageFormat, lambda: obj.photo.height)
        self.assertRaises(InvalidImageFormat, lambda: obj.photo.image)

    @override_app()
    def test_empty_save(self):
        obj = DBDemoModel()
        obj.save()
        obj = DBDemoModel.objects.get(pk=obj.pk)
        self.assertFalse(obj.photo)
        self.assertRaises(InvalidImageFormat, lambda: obj.photo.size)

    @override_app()
    def test_wrong_value(self):
        obj = DBDemoModel(photo='aaaaaaaaa')
        self.assertRaises(InvalidImageFormat, lambda x: obj.photo.width, 0)
        self.assertRaises(InvalidImageFormat, lambda x: obj.photo.height, 0)
        self.assertRaises(InvalidImageFormat, lambda x: obj.photo.size, 0)

    @override_app()
    def test_form(self):
        base64_picture_mime = b64.encodestring(PICTURE)
        Form = modelform_factory(DBDemoModel)
        form = Form({'photo': base64_picture_mime})
        form.is_valid()
        obj = form.save()
        self.assertEqual(obj.photo.size, 53472)
        self.assertEqual(len(obj.photo), 72235)
        self.assertEqual(obj.photo.width, 320)
        self.assertEqual(obj.photo.height, 200)
        self.assertTrue(obj.photo.image)

    @override_app()
    def test_form_error(self):
        Form = modelform_factory(DBDemoModel)
        form = Form({'photo': 111})
        self.assertFalse(form.is_valid())

        form = Form({'photo': '111'})
        self.assertFalse(form.is_valid())


class TestFSCameraField(CameraFieldTest):
    def test_filesystem_create(self):
        shutil.copyfile(PICTURE_PATH, os.path.join(settings.MEDIA_ROOT, PICTURE_NAME))
        base64_picture = b64.encodestring(PICTURE)
        obj = FSDemoModel.objects.create(photo=Base64Image(base64_picture))

        obj.save()

        obj = FSDemoModel.objects.get(pk=obj.pk)
        self.assertEqual(obj.photo.size, 53472)
        self.assertEqual(len(obj.photo), 72235)
        self.assertEqual(obj.photo.width, 320)
        self.assertEqual(obj.photo.height, 200)
        self.assertTrue(obj.photo.image)

    @override_app()
    def test_empty_field(self):
        obj = FSDemoModel()
        obj.save()
        self.assertFalse(obj.photo)
        self.assertRaises(InvalidImageFormat,lambda : obj.photo.size)
        self.assertRaises(InvalidImageFormat,lambda : obj.photo.width)
        self.assertRaises(InvalidImageFormat,lambda : obj.photo.height)
        self.assertRaises(InvalidImageFormat,lambda : obj.photo.image)

    @override_app()
    def test_wrong_value(self):
        obj = FSDemoModel(photo='aaaaaaaaa')
        self.assertRaises(IOError, lambda x: obj.photo.width, 0)
        self.assertRaises(IOError, lambda x: obj.photo.height, 0)
        self.assertRaises(IOError, lambda x: obj.photo.size, 0)



import StringIO
from base64 import decodestring
import base64
import binascii
from PIL import Image
from django.core.files import File
from django.db.models.fields.files import ImageFieldFile, FieldFile
import os


class InvalidImageFormat(Exception):
    pass


class Base64Image(object):
    def __init__(self, stream, format=None, name=None):
        self._base64 = stream
        self.name = name
        self.format = format

    def __len__(self):
        return len(self._base64)

    def __nonzero__(self):
        return bool(self._base64)

    def __str__(self):
        return self._base64

    def __unicode__(self):
        return self._base64

#        def get_mime(self, format):
#            return 'data:image/%s;base64' % format

    def as_base64(self):
        return self._base64

    def _get_size(self):
        return len(self.as_bitmap())

    size = property(_get_size)

    def as_bitmap(self):
        if not hasattr(self, '_bitmap'):
            try:
                content = self._base64
                self._bitmap = decodestring(content)
            except (binascii.Error, TypeError, AttributeError, ValueError) as e:
                raise InvalidImageFormat(e)
        return self._bitmap

    @property
    def image(self):
        if not hasattr(self, '_image'):
            try:
                self._image = Image.open(StringIO.StringIO(self.as_bitmap()))
            except IOError:
                raise InvalidImageFormat(self._base64)
        return self._image

    def _get_width(self):
        return self._get_image_dimensions()[0]

    width = property(_get_width)

    def _get_height(self):
        return self._get_image_dimensions()[1]

    height = property(_get_height)

    def _get_image_dimensions(self):
        if not hasattr(self, '_dimensions_cache'):
            if self.image:
                self._dimensions_cache = self._image.size
            else:
                self._dimensions_cache = (None, None)
        return self._dimensions_cache


#class Base64ImageFieldFile(Base64Image, File):
#    def __init__(self, instance, field, data, name):
#        File.__init__(self, None, name)
#        self.instance = instance
#        self.field = field
#        self.storage = field.storage
#        self._committed = True
#        self._base64= data
#        self._file = None
#        self.name = None
#
#    def __str__(self):
#        return str(self.name)
#
#    def __unicode__(self):
#        return unicode(self.name)
#
#    def __nonzero__(self):
#        return bool(self._base64)
#
#    def as_base64(self):
#        if not self._base64:
#            try:
#                self._base64 = '%s,%s' % (self.get_mime(), base64.encodestring(self.file.read()))
#            except IOError:
#                return None
#        return self._base64
#
#    def delete(self, save=True):
#        # Clear the image dimensions cache
#        if hasattr(self, '_dimensions_cache'):
#            del self._dimensions_cache
#        super(Base64ImageFieldFile, self).delete(save)
#
#    def save(self):
#        self.storage.save(self)
#
#    def _get_file(self):
#        self._require_file()
#        if not hasattr(self, '_file') or self._file is None:
#            self._file = self.storage.open(self.name, 'rb')
#        return self._file
#
#    def _set_file(self, file):
#        self._file = file
#        if file:
#            self.name = file.name
#
#    def _del_file(self):
#        del self._file
#
#    file = property(_get_file, _set_file, _del_file)
#
#    def _require_file(self):
#        if not self._file:
#            raise ValueError("The '%s' attribute has no file associated with it." % self.field.name)
#
#    @property
#    def filename(self):
#        self._require_file()
#        return os.path.basename(self._file.name)



# -*- coding: utf-8 -*-
import tempfile
from django.db import models
from webcam.fields import DBCameraField, FSCameraField
from webcam.storage import CameraFileSystemStorage


class DBDemoModel(models.Model):
    photo = DBCameraField('DatabasePictureField',format='png', null=True, blank=True)

class FSDemoModel(models.Model):
    photo = FSCameraField('FilePictureField',format='jpeg', null=True, blank=True, storage=CameraFileSystemStorage(tempfile.gettempdir()))



import errno
import os
import base64
from django.conf import settings
from django.core.exceptions import SuspiciousOperation
from django.utils._os import safe_join, abspathu


class CameraFileSystemStorage(object):
    def __init__(self, location=None, base_url=None):
        if location is None:
            location = settings.MEDIA_ROOT
        self.base_location = location
        self.location = abspathu(self.base_location)
        if base_url is None:
            base_url = settings.MEDIA_URL
        self.base_url = base_url

    def path(self, name):
        try:
            path = safe_join(self.location, name)
        except ValueError:
            raise SuspiciousOperation("Attempted access to '%s' denied." % name)
        return os.path.normpath(path)

    def file(self, name):
        return file(self.path(name))

    def exists(self, name):
        return os.path.exists(self.path(name))

    def open(self, filename, mode='rb'):
        full_path = self.path(filename)
        return open(full_path, mode)

    def load(self, field, instance):
        try:
            filename = getattr(instance, field.attname)
            with self.open(filename, 'r') as fd:
                return '%s,%s' % (filename.get_mime(), base64.encodestring(fd.read()))
        except IOError:
            return None

    def save(self, attr):
        if attr:
            full_path = self.path(attr.file.name)
            directory = os.path.dirname(full_path)
            if not os.path.exists(directory):
                try:
                    os.makedirs(directory)
                except OSError, e:
                    if e.errno != errno.EEXIST:
                        raise

            if not os.path.isdir(directory):
                raise IOError("%s exists and is not a directory." % directory)

            with open(full_path, "wb") as fd:
                fd.write( attr.as_bitmap() )



import StringIO
from base64 import decodestring
import base64
import uuid
import binascii
from django.db.models import Field
from django.db.models.fields.files import FieldFile, ImageFieldFile
from django.utils.translation import ugettext_lazy as _
from django.utils.text import capfirst
import os
from webcam import forms, widgets
from webcam.storage import CameraFileSystemStorage
from PIL import Image
from webcam.utils import InvalidImageFormat, Base64Image

class CameraPictureDescriptor(object):

    def __init__(self, field):
        self.field = field

    def __get__(self, instance=None, owner=None):
        if instance is None:
            raise AttributeError(
                "The '%s' attribute can only be accessed from %s instances."
                % (self.field.name, owner.__name__))
        data_or_filename = instance.__dict__[self.field.name]

        if isinstance(data_or_filename, basestring) or data_or_filename is None:
            if isinstance(self.field, FSCameraField):
                if data_or_filename:
                    with self.field.storage.open(data_or_filename, 'r') as fd:
                        c = fd.read()
                        value = base64.encodestring(c)
                else:
                    value = None
                attr = self.field.attr_class(value, name=data_or_filename)
            else:
                attr = self.field.attr_class(data_or_filename)
        elif isinstance(data_or_filename, (Base64Image, )):
            attr = data_or_filename

        instance.__dict__[self.field.name] = attr

        return instance.__dict__[self.field.name]

    def __set__(self, instance, value):
        instance.__dict__[self.field.name] = value

class DBCameraPictureDescriptor(CameraPictureDescriptor):
    pass

class FSCameraPictureDescriptor(CameraPictureDescriptor):
    pass

class CameraField(Field):
    description = _("Picture")

    descriptor_class = CameraPictureDescriptor

    def __init__(self, verbose_name=None, name=None, format='jpg', widget_width=320, widget_height=240,
                 width_field=None, height_field=None, **kwargs):
        for arg in ('primary_key', 'unique'):
            if arg in kwargs:
                raise TypeError("'%s' is not a valid argument for %s." % (arg, self.__class__))
        if format not in ['gif', 'jpeg', 'png']:
            raise TypeError("'%s' is not a valid format. ([gif|jpg|png])" % format)

        self.format = format
        self.widget_width = widget_width
        self.widget_height = widget_height
        self.width_field, self.height_field = width_field, height_field
        super(CameraField, self).__init__(verbose_name, name, **kwargs)


    def formfield(self, **kwargs):
        defaults = {'form_class': forms.CameraField,
                    'format': self.format,
                    'width': self.widget_width,
                    'height': self.widget_height,
                    'required': not self.blank,
                    'label': capfirst(self.verbose_name),
                    'help_text': self.help_text}
        defaults.update(kwargs)
        return super(CameraField, self).formfield(**defaults)

    def update_dimension_fields(self, instance, force=False, *args, **kwargs):
        has_dimension_fields = self.width_field or self.height_field
        if not has_dimension_fields:
            return

    def contribute_to_class(self, cls, name):
        super(CameraField, self).contribute_to_class(cls, name)
        setattr(cls, self.name, self.descriptor_class(self))

#    def get_mime(self, format):
#        return 'data:image/%s;base64' % format

class DBCameraField(CameraField):
    """
        Camera field that store the picture into the database
    """
    attr_class = Base64Image

    def pre_save(self, model_instance, add):
        attr = getattr(model_instance, self.attname)
        return attr.as_base64()

    def get_internal_type(self):
        return "TextField"

    def formfield(self, **kwargs):
        kwargs['widget'] = widgets.DBCameraWidget
        return super(DBCameraField, self).formfield(**kwargs)


class FSCameraField(CameraField):
    """
        Camera field that store the picture as file just like ImageField
    """
    attr_class = Base64Image

    def __init__(self, verbose_name=None, name=None, format='jpg', widget_width=320, widget_height=240,
                 width_field=None, height_field=None, storage=None, **kwargs):
        self.storage = storage or CameraFileSystemStorage()
        super(FSCameraField, self).__init__(verbose_name, name, format, widget_width, widget_height,
                                            width_field, height_field, max_length=100, **kwargs)

    def get_internal_type(self):
        return "CharField"

    def value_from_object(self, obj):
        return self.storage.load(self, obj) or ""

    def get_prep_value(self, value):
        """
        Perform preliminary non-db specific value checks and conversions.
        """
        if isinstance(value, self.attr_class):
            return str(value)
        return value

    def bind(self, fieldmapping, original, bound_field_class):
        return bound_field_class(self, fieldmapping, original)

    def pre_save(self, model_instance, add):
        attr = getattr(model_instance, self.attname)
        if attr:
            filename = self.get_filename(model_instance)
            if add:
                attr.file = self.storage.open(filename, 'wb')
                attr.name = filename
            self.storage.save(attr)
            return filename
        return None

    def get_filename(self, instance):
        return str(getattr(instance, self.name).name or "%s.%s" % (str(uuid.uuid4()), self.format))

    def save_form_data(self, instance, data):
        filename = self.get_filename(instance)
        attr = getattr(instance, self.attname)
        setattr(instance, self.name, filename)

    def formfield(self, **kwargs):
        kwargs['widget'] = widgets.FSCameraWidget
        return super(FSCameraField, self).formfield(**kwargs)

    def to_python(self, value):
        """
        Converts the input value into the expected Python data type, raising
        django.core.exceptions.ValidationError if the data can't be converted.
        Returns the converted value. Subclasses should override this.
        """
        return value

    def _get_val_from_obj(self, obj):
        if obj is not None:
            return getattr(obj, self.attname)
        else:
            return self.get_default()

try:
    import south
    from south.modelsinspector import add_introspection_rules

    add_introspection_rules([], ["^webcam\.fields\.DBCameraField"])
    add_introspection_rules([], ["^webcam\.fields\.FSCameraField"])
except ImportError:
    pass


